/*
 * V11.C
 *
 *      Center line module.
 *      Line reform module.
 *      Help routines for LOADINI option.
 */
#include "v.h"

/*
 * REFORM
 *
 *      The purpose of this subroutine is to act on a given line pair.
 *      These actions are taken:
 *
 *        1. Processing starts at the current cursor position.
 *        2. Leading spaces are replaced by the number of spaces
 *           in the current tab setting unless INDENAUTO!=0. See 8.
 *        3. Spaces after the leading spaces are reduced to a
 *           single space, except for ".!?" followed by two or
 *           more spaces, which are reduced to two spaces.
 *        4. If the line becomes full as a result of copying
 *           characters, then the line is split into two lines.
 *        5. If the line does not fill, then the characters on the
 *           next line are used to help fill the line. If that
 *           line is completely blank, then processing stops.
 *        6. At most one line back or ahead is accessed.
 *        7. The cursor remains at the margin column of the current
 *           line, or it is placed at the start of the next line if
 *           that line has characters other spaces, or it is placed at
 *           the end of the line, in case the next line is blank.
 *        8. The indent rules are fairly complex. The idea is to do an
 *           indent differently in col==0 than for a col>0. If col==0,
 *           then indent happens. Otherwise, the indent has no effect.
 *           The amount of indent depends upon the setting of variable
 *           INDENAUTO:
 *             case 1. INDENAUTO==0
 *             Then indent=0, if there is no white space at the front of
 *             the line. Otherwise indent=TSTOPS.
 *             case 2. INDENT != 0
 *             The indent will equal the indent of the previous line, if
 *             that makes sense, otherwise TSTOPS is used, unless the
 *             current line is without leading space, in which case the
 *             indent is zero.
 *        9. Reform from a blank line onto a full line will go to the
 *           start of the full line. This prevents repeat keys from
 *           corrupting a previously reformed paragraph. The effect is
 *           undone by INDENAUTO==0.
 */
#if !HASSOB
char *sob(p) char *p; {
  while(p[0] == ' ') ++p;
  return p;
}
#endif

#if !HASFNB
char *fnb(p) char *p; {
  while(p[0] != EOS && p[0] != ' ') ++p;
  return p;
}
#endif

/* center text within margins */
void textcenter(){
int i; char *p; char *sob();
  strip(p=sob(line[row]));                      /* trim off blanks */
  if((i=strlen(p)) <= margin){
    i=(margin-i)/2;                             /* add leading blanks */
    fillchr(sysvv,' ',i);                       /* to target string */
    strcpy(&sysvv[i],p);                        /* copy tail */
    fixline();                                  /* fixup screen */
    eddn(); curses();
  } else bell();
}

void reform() {
char *r;
int ind;
char safe[512];
int oldcol;
extern char *ju(),*sob();
extern unsigned char *curchar();
extern char *sysvv;
extern int firstnonblank();

  if(ROWMAX <= 1) return;
/* A mistake? Need to add curoff support someday. */
  if(curoff>0){setoffset(curoff=0); drawall();}
  curchar();                                 /* Add blanks as needed. */
  ind=(!col && line[row][0]==' ') ? TSTOPS:0;/* default indent */
  if(INDENAUTO){                             /* set indent for AUTO */
    if(col) ind=firstnonblank(line[row]);    /* use line's indent */
    else if(lncount>1) {
      edup(); eddn(); curses();              /* on row 2 or greater */
      if(*sob(r=line[row-1]))                /* Use indent of */
        ind=firstnonblank(r);                /* previous line. */
    }
    /* Otherwise, use the default indent already set. */
  }
  if(row == ROWMAX) { eddn(); edup(); curses();}
  if(*sob(line[row]) == EOS) {               /* Line all blanks */
    eddn(); col=firstnonblank(line[row]); goto quit;
  }
  if(*sob(line[row+1]) == EOS && *sob(curchar()) == EOS)
    goto nextline;                           /* At end, next line blank */
  if(ind>=margin) goto nextline;             /* Indent too large */
  /* Search the line for a word boundary on which to begin reform. */
  r=line[row];
  oldcol=col;
  while(col<margin && r[col] && r[col] != ' ') ++col;
  if(r[col] && r[col] != ' '){
    col=min(col,margin);
    while(col>0 && r[col] != ' ') --col;
    while(r[col] == ' ' && col>0 && r[col-1] == ' ') --col;
    if(col==0) {bell(); goto nextline;}
  }
  strnzcpy(sysvv,r,col);                     /* Copy up to cursor */
  if(!col) fillfront(sysvv,ind,"");          /* Nothing? Then indent.*/
  r = ju(sysvv,r+col,margin);                /* Justify into sysvv[] */
  if(r[0]) {                                 /* from cursor forward.*/
    if(INDENAUTO){
      if(ind+strlen(r)<MAXLN) fillfront(safe,ind,r);
      else { fixline(); goto setcursor; }
    } else
    strcpy(safe,r);
    fixline();
    eddn(); curses();
    addln(safe,TRUE);
    col = firstnonblank(safe); goto quit;
  }
  else {
    r = line[row+1];
    if(*sob(r)==0 || (!INDENAUTO && (r[0] == EOS || r[0] == ' '))) {
       /* new paragraph */
       if(oldcol<strlen(line[row])){fixline(); goto setcursor;}
       goto nextline;
    }
    else {                              /* next line has chars */
      r = ju(sysvv,r,margin);           /* add more onto sysvv[] */
      fixline();                        /* print what we got */
      /* now do the next line, from the spillover */
      if(INDENAUTO){
        if(ind+strlen(r) < MAXLN) fillfront(sysvv,ind,r);
        else goto setcursor;
      } else
      strcpy(sysvv,r);
      eddn(); curses();
      if(r[0]) {fixline(); col = firstnonblank(sysvv); goto quit;}
      else { delln(sysvv); edup();}
    }
  }
goto setcursor;
nextline:
  eddn();
setcursor:
  col = min(strlen(line[row]),margin);
  CenterRefresh();      /* if on last line then refresh to center */
quit:
  return;
}

/*
 * FIXLIN
 *
 * Purpose:
 *              Erases the current line, replaces it with the
 *              line held in sysvv[].
 */
void fixline()
{
extern char *toprint();
  xPrint(toprint(sysvv),row,col=0); eraseol();
  strcpy(line[row],sysvv);
}

/*
 * FIRSTNONBLANK
 *
 * Find string index to first nonblank char or to end of string
 */
int firstnonblank(s) char *s; {
extern char *sob();
  return (int)(sob(s)-s);
}

/*
 * FILLFRONT
 *
 * Fills a string with blanks then appends another string.
 *
 */
void fillfront(d,ind,s) char *d,*s; int ind; {
  fillchr(d,' ',ind); strcpy(d+ind,s);
}

/*
 * JU
 *
 * Purpose: Does single line justification.
 *
 The basic algorithm is to begin this routine with sysvv[]
 filled. The contents of array buf[] are added to the end of
 sysvv[] until the margin is exceeded. The position in buf[]
next to be processed is returned (hopefully end of string).

 */
char *ju(d,s,margin)
char *d;        /* destination buffer */
char *s;        /* source buffer */
int margin;     /* right margin limit */
{
char *r,*p;
int i,j,flag;
char *sob(),*fnb(),*strchr();

  j = strlen(d);
  p = fnb(r=sob(s));
  while(r[0]) {
    if((i = (int)((fnb(r))-r)) == 0) break;
    flag = (j > 0 && d[j-1] != ' ');
    if(flag){
      ++i;
      if(strchr(".!?",d[j-1])!=(char *)0) {  /* Worcester add on 1992 */
        if(p[0]==' ' && p[1]==' ') d[j++]=' ';
      }
    }
    if((i+j) <= margin) {
      if(flag) { d[j++] = ' '; --i; }   /* add space after token*/
      while(i--) d[j++] = *r++;         /* copy token */
      r = sob(p=r);                     /* position to next token */
    } else break;
  }
  d[j] = EOS;
  return r;
}

#if PIHELPDOC
/* PIHELPH.C  - color or mono version of pihelp for HTML source     */
extern long ftell();

#define SLASHCHAR '/'
void setcolors(){}
void setcolormono(){}
void setcolorinverse(){}
void cputs(s) char *s; { ePrint(s);}
void clrscr(){ clearscreen(); home();}

void
cr(){ etype("\r"); }

int isHTML(s) char *s; {
  if(s[0]=='<') return 1;
  return 0;
}

/* Detect HTML line  <a NAME="blah blah blah"></a> */
int isaheader(s) char *s; {
char *p;
extern char *strstr(),*strchr();
char t[256];
  if(! isHTML(s) ) return 0;
  if(strchr("aA",s[1])==0) return 0;
  if(strstr(s,"name=")!=0 || strstr(s,"NAME=")!=0){
    p=strchr(s,'=');
    if(p==0) return 0;
    p=strchr(s,'"');
    if(p[1]=='"') ++p;
    if(p==0) return 0;
    ++p;
    strcpy(t,p);
    p=strchr(t,'"');
    if(p==0) return 0;
    p[0]=0;
    strcpy(s,t);
    return 1;     /* Header title in string s[] */
  }
  return 0;
}

void
getReturnKey(){
  pressany("");
}

void
killLinefeed(s) char *s; {
extern char *strchr();
char *p;
  if((p=strchr(s,'\n'))!=(char *)0) p[0]=0;
  if((p=strchr(s,'\r'))!=(char *)0) p[0]=0;
}

int fgetline(s,fp) char *s; FILE *fp; {
  if(fgets(s,128,fp)==(char *)0) return 0;
  killLinefeed(s);
  return 1;
}

#define ROWS 8          /* Use 8 rows of items per help screen */
#define MAXSTRING 256   /* Largest string expected in helpfile */
#define MAXLPS 22       /* Max lines per screen */

#define INFILE "pihelp.htm"     /* Name of the helpfile */
#define TBLSIZ 500              /* Max number of index items */


void help() {
long *table;
char **token;
char *p,*q;
FILE *fp;
char file[128];
char s[MAXSTRING],keep[MAXSTRING];
long place;
int i,n,j,k,m,x,base,nlines;
int looping,maxlines;
char header[128];
char leader[128];
extern char *strchr(),*strrchr();
extern char *sobl();
extern char *PIHLPDOC;

  if(ROWMAX<13 || COLMAX<72){
    pressany("Screen size too small for manual");
    return;
  }

  if((table =(long *)malloc(sizeof(long)*TBLSIZ))==(long *)0) goto bad;
  if((token =(char **)malloc(sizeof(char *)*TBLSIZ))==(char **)0){
     free(table); goto bad;
  }

  strcpy(file,PIHLPDOC);
  fp = fopenbread(file);
  if(fp == (FILE *)0) {
   cputs("Can't open file "); cputs(file); crlf();
   goto bad;
  }

  maxlines = MAXLPS;
  setcolors();

  s[0]=n=0;
  while(n<TBLSIZ) {
    place = (long)ftell(fp);
    if(fgetline(s,fp) <= 0) break;
    if(isaheader(s)){                  /* Changes s[] to header text */
      if(!safecore(&token[n],s)) break;
      table[n++] = place;
    }
  }

  base = 0;
  looping = 1;
  while(looping) {

   clrscr();
   m = (n < base+ROWS) ? n : (base+ROWS);
   k = n/ROWS; if((n%ROWS)!=0) ++k;
   sprintf(header,"%s %d of %d","        HELP TOPICS",1+base/ROWS,k);
   cputs(header); crlf(); crlf();
   for(k=0,i=base;i<m;++i,++k) {
      sprintf(leader,"        %d  -  ",1+k);
      cputs(leader);
      cputs(token[i]); crlf();
   }
   crlf();
   cputs("        N  -  Next Screen");
   crlf();
   cputs("        P  -  Previous Screen");
   crlf();
   cputs("        Q  -  Quit");
   crlf();
   crlf();
   cputs("Select: ");

   x = getupper();
   switch(x) {
   case '1': case '2': case '3': case '4':
   case '5': case '6': case '7': case '8':
             j = x - '1' + base; break;
   case 0: case ' ': case '\r': case '\n': case 'N': case 'N'-'@':
             base = base + ROWS; if(base >= n) base = 0;
             j = -1; break;
   case 127: case 8: case 'B': case 'P': case 'P'-'@':
             base = base - ROWS; if(base < 0) { base = n - n%ROWS;}
             if(base < 0) base = 0;
             j = -1; break;
   case 3:   case 'X': case 'Q':
             looping = 0; j = -1; break;
   default: bell(); j = -1; break;
   }
repeat:
   if(0 <= j && j < n) {
      clrscr();
      s[0]=nlines=0;
      fseek(fp,table[j],0);
      if(fgetline(s,fp) <=0) goto badseek;
      if(!isaheader(s)) goto badseek;
      strinvvideo(s); crlf();
      strcpy(keep,s);
      while(1) {
          if(fgetline(s,fp) <= 0) break;
          if(isaheader(s)){
            while(nlines++ < maxlines) crlf();
            break;
          }
          if(isHTML(s)) continue;
          cputs(s); crlf(); ++nlines;
          if(nlines >= maxlines) {
             cputs("--MORE--"); x = getupper();
             if(strchr("QXESN\3\32",x) != (char *)0) break;
             clrscr();
             strinvvideo(keep); crlf();
             nlines=1;
          }
      }
      cr(); cputs("END OF SECTION. Press Q to quit: ");
      if(strchr("QqXxSs\3\32",getupper())==(char *)0){++j; goto repeat;}
   }
  }

stop:
  closedfile(fp);
  goto normalexit;
badseek:
 cputs("Bad seek in "); cputs(file); crlf();
 getReturnKey();
normalexit:
  for(i=0;i<n;++i) free(token[i]);
  free(token);
  free(table);
normal1:
  setattribute(0);
  return;
bad:
  getReturnKey();
  goto normal1;
}

#endif /* end PIHELPDOC */
