/*
 * V3.C
 *
 */

#include "v.h"

/*
 * STILLR
 *
 * Purpose:
 *              Test for room to push line onto low
 *              or high buffer. Need HEADROOM bytes free.
 * Returns:
 *              TRUE  - if room is available
 *              FALSE - out of room
 *                      Prints NO ROOM message on banner line
 * Notes:
 *              o Lines are up to 1+MAXLN long. Don't let the
 *                editor operate with less than 1+MAXLN bytes free.
 *              o This module does a lot of the dirty work for
 *                the 640k memory model. It silently purges the
 *                edit buffer when nearly full, so that attempts
 *                to add strings will succeed.
 */
int stillroom() {
  if(HEADROOM< hbuf-lbuf) return TRUE;
  /* might have to purge 2k from the buffer */
  if(bufbase+SIZEBLOCK <= lbuf) putLOWblock();
  else
  if(SIZEBLOCK<=ebuf-hbuf) putHIblock();
  return ( (HEADROOM< hbuf-lbuf) ? TRUE : noroom());
}

/*
 * LOPUSH
 *
 * Purpose:
 *              Push a line p onto the low buffer
 * Returns:
 *              p     - success
 *              NULLP - failure
 * Notes:
 *              o Strip blanks
 *              o update lbuf
 *              o check for available room
 */
char *loPush(p) char *p; {
int strlen();

  if(stillroom()) {
    if(curfile && BLANKSTRIP) strip(p);
    strcpy(lbuf+1,p);
    lbuf += strlen(p)+1;
    return p;
  }
  return NULLP;
}

/*
 * HIPUSH
 *
 * Purpose:
 *              Push a line onto the high buffer.
 * Returns:
 *              p     - if successful
 *              NULLP - on failure
 * Notes:
 *              o Check for room.
 *              o update hbuf, the high buffer pointer
 */
char *hiPush(p) char *p; {
int strlen();

  if(stillroom()) {
    if(curfile && BLANKSTRIP) strip(p);
    if(p[0]==0){
      /* don't add blank lines onto the very end of the buffer */
      if(hbuf>=ebuf && finfo[curfile].nbTOT<=finfo[curfile].nbLOW)
        return p;
    }
    hbuf -= strlen(p)+1;
    strcpy(hbuf,p);
    return p;
  }
  return NULLP;
}

/*
 * LOPULL
 *
 * Purpose:
 *              Get a line from the low buffer.
 * Returns:
 *              TRUE  - it worked
 *              FALSE - it failed
 *              lbuf  - updated
 */
int loPull(p) char *p; {
char *strstart(),*endof(); char *detab();
char *q,*r;
  if(lbuf<bufbase){
    if(getLOWblock()!=0) return FALSE;
    if(hbuf-lbuf <= HEADROOM) putHIblock();
  }
  if(lbuf<bufbase) return FALSE;
  *lbuf=0;
#if 0 /* Remove stripping of CR in CRLF type files May 1995 */
  if(intexttype==1 && lbuf[-1] == 13) *--lbuf = 0;
#endif
  lbuf = strstart(r=lbuf);
  if(DETABBER && curfile && dotabs(lbuf+1)) {
    q=detab(lbuf+1,p);
    if(q<r) {strcpy(lbuf+1,q); lbuf = endof(lbuf+1);}
  }
  else
  strcpy(p,lbuf+1);
  if(BITSTRIP) unsetbits(p);
  return TRUE;
}

/*
 * HIPULL
 *
 * Purpose:
 *              Pull a line off the high buffer.
 * Returns:
 *              TRUE  - it worked
 *              FALSE - it failed
 * Notes:
 *              o update hbuf
 *              o don't do it if high buffer is empty
 */
int hiPull(p) char *p; {
int strlen(); char *detab();
  if(hbuf>=ebuf) {
    if(getHIblock()==0){
      if(HEADROOM >= hbuf-lbuf) putLOWblock();
    }
  }
  if(hbuf>=ebuf) { /* add blank lines as requested */
    *p = EOS;
    return TRUE;
  }
  if(DETABBER && curfile && dotabs(hbuf)) {
    hbuf=detab(hbuf,p);
#if 0 /* Remove stripping of CR in CRLF type files May 1995 */
    if(intexttype==1 && (i=strlen(p))>0 && p[i-1]==13) p[i-1]=0;
#endif
  }
  else {
    strcpy(p,hbuf);
    hbuf += strlen(p)+1;
#if 0 /* Remove stripping of CR in CRLF type files May 1995 */
    i=strlen(p);
    if(intexttype==1 && i>0 && p[i-1]==13) p[i-1]=0;
#endif
  }
  if(hbuf>ebuf) hbuf = ebuf;
  if(BITSTRIP) unsetbits(p);
  return TRUE;
}

void decrROW(){
 --row; --lncount;
}

void incrROW(){
 ++row; ++lncount;
}

/*
 * LINEUP
 *
 * Purpose:
 *              Scroll up one line.
 * Notes:
 *              o Moves last line onto high buffer
 *              o Pulls new line from low buffer
 *              o update screen, pointers
 *              o only works if low buffer not empty
 *              o If out of room, then last line is
 *                thrown away, with warning banner.
 */
void lineup() {
int i;
char *p;
char *hiPush(),*toprint();

  if(bufbase > lbuf) getLOWblock();
  if(bufbase <= lbuf) {
      if(hiPush(p=line[ROWMAX]) == NULLP) {
              bellpressany("NO ROOM - Last line lost"); return;
      }
      revindex();
      loPull(p);
      for(i=ROWMAX-1;i>=0;--i) line[i+1] = line[i];
      line[0]=p;
      if(lncount) --lncount;
      xPrint(toprint(p),0,0);
  }
}

/*
 * LINEDN
 *
 * Purpose:
 *              Scroll off the first line, scroll in
 *              a new last line.
 * Notes:
 *              o Move in a blank line if the high buffer is empty
 *              o Fails if the low buffer push fails
 *              o Screen update indicates success
 */
void linedn() {
int i;
char *p;
char *loPush(),*toprint();
int strlen();

  if((p = loPush(line[0])) != NULLP) {
    fwdindex();
    for(i=0;i<ROWMAX;++i) line[i] = line[i+1];
    hiPull(line[ROWMAX]=p);
    xPrint(toprint(p),ROWMAX,0);
    ++lncount;
  }
}


/*
 * REFRESHH
 *
 * Purpose:
 *              Refresh(off) computes the target line as the current
 *              line plus off and brings it to the top of the screen.
 *              Refresh is either NEAR (smooth scroll) or FAR (full
 *              screen clear plus print all lines). We define FAR as
 *              more than 2 screens away (backward or forward).
 * Returns:
 *              TRUE  - Refresh done as requested.
 *                      lncount = target line.
 *                      row & col unchanged.
 *              FALSE - Refresh failed due to no room.
 * Notes:
 *              o This was the most difficult code to write in the
 *                entire editor. Change it with caution.
 *              o lncount -  current line number (1...maxint)
 *              o hiPush(), hiPull(), loPush(), loPull(), lineup(),
 *                linedn() are previously defined editor functions.
 *              o Out-of-range requests are truncated to the nearest
 *                reasonable value.
 */
int Crefreshh(off) LLONG off; {   /* Refresh with cursor center screen */
int newrow; int r;
extern int row; extern LLONG lncount;
  newrow=middle((LLONG)off);
  r=((off=refreshh(off-newrow))!=0);
  if(r) lncount += (row=newrow);
  return r;  /* 0 if off==0 and 1 otherwise */
}

void Prefreshh(dir) int dir; {   /* Refresh to next page, retain cursor
*/
int i;                      /* Argument is in array sysv[] */
extern LLONG atoiLL();
  lncount -= (i = row); row=0;
  refreshh(dir*atoiLL(sysv)*(1+ROWMAX));
  lncount += (row=i);
}

int refreshh(off) LLONG off; {
LLONG target,offset,top,bot;
int tmp,flag;
int PackScreenBuffer();
int loPull(),hiPull();

  target = lncount+off;
  if(off < 0 && lncount <= -off) target = 1;
/*
 *      Compute line numbers for top and bottom of screen
 */
  top = lncount - row;
  bot = top + ROWMAX;
  flag = FALSE;
  deFRAG();  /*insure low and high buffers do not fragment */
/*
 *      Is target on current screen?
 *      Then scroll target to top line.
 */
  if(top <= target && target <= bot) {
    offset = target - top;
    while(offset-- > 0) linedn();
  }
  else
/*
 *      Is target above current screen?
 *      Then test for FAR or NEAR scroll.
 */
  if(target < top) {
    offset = top - target;
    if(offset < FARSCROLL) {            /* NEAR scroll */
      PUSHnodisp();
      if(!(HASINSERTLINE+HASSCROLL)) nodisp = flag = 1;
      while(offset-- > 0) lineup();
      POPnodisp();
    }
    else {                                  /* FAR scroll */
      if(!PackScreenBuffer(0)) {
        offset -= ROWMAX+1;
        while(offset-- > 0) {
          if(!loPull(sysx) || !hiPush(sysx)) {
            target += 1+offset;         /* correct it */
            break;
          }
        }
        UnPackScreenBuffer(ROWMAX+1);
        flag = TRUE;
      } else return FALSE;
    }
  }
  else
/*
 *      Is target below current screen?
 *      Then test for FAR or NEAR scroll.
 */
  if(target > bot) {
    offset = target - bot + ROWMAX;
    if(offset < FARSCROLL) {            /* NEAR scroll */
      while(offset-- > 0) linedn();
    }
    else {                              /* FAR scroll */
      if(!PackScreenBuffer(ROWMAX+1)) {
        offset -= 1+ROWMAX;
        while(offset-- > 0) {
          if(!hiPull(sysx) || !loPush(sysx)){
            target -= 1+offset;         /* correct it */
            break;
          }
        }
        UnPackScreenBuffer(0);
        flag = TRUE;
      } else return FALSE;
    }
  }
  lncount = target;
  if(flag) drawall();
  return TRUE;
}

/*
 * EDUP
 *
 * Purpose:
 *              Move position on screen up one line.
 * Notes:
 *              o If at top of screen, then pull in a line.
 *              o Keep track of line count.
 *              o Display line number and free space.
 */
void edup() {
  if(row) decrROW();
  else
  lineup();
}

/*
 * EDDN
 *
 * Purpose:
 *              Move cursor down one line.
 * Notes:
 *              o Scroll down as needed
 *              o Update line count
 *              o Display line count when done
 */
void eddn() {
  if(row < ROWMAX) incrROW();
  else
  linedn();
}

/*
 * HOMEIT
 *
 * Put screen back to left edge.
 *
 */
void homeit(){
  col = 0;
  if(curoff){curoff=0; drawall();}
}

/*
 * ENDIT
 *
 * Put screen back to right edge.
 *
 */
void endit(){
   col = strlen(line[row]);
   if(col>curoff+COLMAX || col<curoff) curfix(1+col,0);
   else col -= curoff;
}

/*
 * EDRIGH
 *
 * Purpose:
 *              Move cursor right one space
 */
void edright() {
#define JUMPRIGHT 5
if(col < COLMAX) ++col;
else {
  if(col+curoff<MAXLN-1){
    curoff=min(curoff+JUMPRIGHT,MAXLN-1-col);
    col=max(0,col+1-JUMPRIGHT);
    setoffset(curoff);
    drawall();
  }
}
}

/*
 * EDLEFT
 *
 * Purpose:
 *              Move cursor left one space.
 */
void edleft() {
#define JUMPLEFT 5
if(col) --col;
else { /* in col=0 */
  if(curoff>0) {
    curoff=max(curoff-JUMPLEFT,0);
    col=min(COLMAX,col+JUMPLEFT-1);
    setoffset(curoff); drawall();
  }
}
}

/*
 * ADDLN
 *
 * Purpose:
 *              Add a blank line to the screen. Push last line
 *              on screen to high buffer.
 * Returns:
 *              TRUE  - line was added
 *              FALSE - failed to add line due to no room.
 */
int addln(q,flag) char *q; int flag; {
int i;
char *p;
char *hiPush(),*toprint();

  if((hiPush(p=line[ROWMAX])) != NULLP) {
    if(flag) lineinsert();
    for(i=ROWMAX-1;i>=row;--i) line[i+1] = line[i];
    strcpy(line[row]=p,q);
    if(flag) xPrint(toprint(p),row,0);
    return TRUE;
  }
  return noroom(); /* noroom() returns FALSE */
}

/*
 * ADDSTRING
 *
 * Purpose:
 *              Add a string to the screen at the cursor position.
 *              The destination d[] is assumed a screen line
 *              or buffer of physical buffer size MAXLN.
 *              Position = [row,col] at offset=curoff
 * Returns:
 *              k  - string was added of length k>=0
 *             -1  - failed to add string because of no room.
 */
int addstring(s,d,flag,insmode)
char *s; char *d; int flag,insmode; /* s=source, d=destination, flag=1 or 0 */
{
int i,j,k,l;

  k=strlen(s);
  j=strlen(d);
  i=curoff+col;
  while(j<i) d[j++]= ' ';
  d[j]=0;
  l=(insmode ? (j+k): max(j,i+k));
  if(l<MAXLN-1){
    d += i;
    if(insmode) strcpy(sysv,d);
    moveMEM(d,s,k);
    if(insmode) strcpy(d+k,sysv);
    d[l-i]=0;
    if(flag) xPrint(d,row,col);
    return k;
  }
  bell();
  return -1;
}

/*
 * DELLN
 *
 * Purpose:
 *              Delete a line on the screen from the screen
 *              buffer, and hence from the main buffer.
 * Returns:
 *              Deleted line copied to string pointer r.
 */
void delln(r) char *r; {
int i;
char *p;
extern char *toprint();

  linedelete();
  strcpy(r,p = line[row]);
  for(i=row;i<ROWMAX;++i) line[i] = line[i+1];
  hiPull(line[ROWMAX]=p);
  xPrint(toprint(p),ROWMAX,0);
  curses();
}
